"""Tests for interactors.retrieve module."""
import pytest
from bacore.domain import settings
from bacore.interactors import retrieve


@pytest.fixture
def fixture_test_system_information():
    """Fixture for system_information_os."""
    return "Darwin"


def test_system_information_os(fixture_test_system_information):
    """Test system_information_os."""
    information = retrieve.system_information_os(func_os=fixture_test_system_information)
    assert isinstance(information, settings.System)
    assert information.os == "Darwin"



"""Top level conftest.py for BACore test cases."""
import pytest


@pytest.fixture
def fixture_pyproject_file(tmp_path):
    """Create a temporary pyproject.toml file."""
    toml_content = """
    [project]
    name = 'bacore'
    version = "1.0.0"
    description = "BACore is a framework for business analysis and test automation."
    """
    file = tmp_path / "pyproject.toml"
    file.write_text(toml_content)
    return file



"""Retrieve Functionality Module (the "get" word feels overloaded)."""
import platform
from bacore.domain import settings
from typing import Protocol, runtime_checkable


@runtime_checkable
class SupportsRetrieveDict(Protocol):
    """Protocol for retrieval of file content as dict."""

    def data_to_dict(self) -> dict:
        """Content as dictionary."""
        ...


def file_as_dict(file: SupportsRetrieveDict) -> dict:
    """Content as dictionary."""
    return file.data_to_dict()


def system_information_os(func_os: callable = platform.system()) -> settings.System:
    """Retrieve system information."""
    information = settings.System(os=func_os)
    return information



"""Module domain files for handling of files and directories."""
import tomllib
from pathlib import Path


class TOML:
    """TOML file class."""

    def __init__(self, path: Path):
        """Initialize."""
        self.path = path

    @property
    def path(self):
        """Get file path."""
        return self._path

    @path.setter
    def path(self, value):
        """Set file path as pathlib.Path object."""
        if not isinstance(value, Path):
            raise TypeError("Path must be a pathlib.Path object.")
        self._path = value

    def data_to_dict(self) -> dict:
        """Content as dictionary."""
        with open(self.path, mode="rb") as f:
            content = tomllib.load(f)
        return content



"""Measurements module."""
from datetime import date, datetime, timedelta
from dataclasses import dataclass


@dataclass
class Measurement:
    """Measurements class which contains measurements related settings.

    Attributes:
        name: The name of the measurement.
        unit: The unit of the measurement.
        value: The value of the measurement.
    """

    name: str
    unit: str
    value: float


@dataclass(frozen=True)
class Time:
    """Class for time related settings."""

    now = datetime.now()
    now_s = now.strftime("%Y-%m-%d %H:%M:%S")
    today = date.today()
    today_s = today.strftime("%Y-%m-%d")
    yesterday = today - timedelta(days=1)
    yesterday_s = yesterday.strftime("%Y-%m-%d")
    ty = today.strftime("%Y")
    tm = today.strftime("%m")
    td = today.strftime("%d")
    yy = yesterday.strftime("%Y")
    ym = yesterday.strftime("%m")
    yd = yesterday.strftime("%d")



"""Configuration module for BACore."""

from bacore.domain import settings

project_bacore = settings.Project(name="BACore",
                                  version="0.0.10",
                                  description="BACore is a framework for business analysis and test automation.")



"""Tests for interactors.verify module."""
import pytest
from bacore.domain import system
from bacore.interactors import retrieve, verify

pytestmark = pytest.mark.interactors


@pytest.fixture
def fixture_test_command_on_path():
    """Fixture for command_on_path."""
    if retrieve.system_information_os().os in ["Darwin", "Linux"]:
        return "ls"
    else:
        return "dir"


def test_command_on_path(fixture_test_command_on_path):
    """Test command_on_path."""
    command = system.CommandLineProgram(name=fixture_test_command_on_path)
    assert verify.command_on_path(command=command) is True
    bogus_command = system.CommandLineProgram(name="bogus_does_not_exist")
    assert verify.command_on_path(command=bogus_command) is False



"""Settings module for settings of BACore and its components."""
import platform
from bacore.domain import files
from pathlib import Path
from pydantic import Field, computed_field, field_validator, SecretStr
from pydantic_settings import BaseSettings
from pydantic.dataclasses import dataclass
from typing import Optional


class Credential(BaseSettings):
    """Credential details."""

    username: str
    password: SecretStr

    @field_validator("username")
    @classmethod
    def username_must_not_contain_spaces(cls, v: str) -> str:
        """Validate that the username does not contain spaces."""
        if " " in v:
            raise ValueError("No spaces allowed in username.")
        return v


@dataclass
class Project:
    """Project information."""

    name: str
    version: Optional[str] = None
    description: Optional[str] = None

    @field_validator("name")
    @classmethod
    def name_must_not_contain_spaces(cls, v: str) -> str:
        """Validate that the name does not contain spaces."""
        if " " in v:
            raise ValueError("No spaces allowed in project name.")
        return v


class ProjectSettings(BaseSettings):
    """Project settings assumes that the project will use a pyproject.toml file."""

    path: Path = Field(default=Path("."), alias="project_root_dir")

    @field_validator("path")
    @classmethod
    def path_must_be_directory(cls, v: Path) -> Path:
        """Validate that the path is a directory."""
        if v.is_dir() is False:
            raise ValueError(f"Path '{v}' is not a directory.")
        return v

    @computed_field
    @property
    def _pyproject_file(self) -> Path:
        project_file = self.path / "pyproject.toml"
        if project_file.is_file() is False:
            raise FileNotFoundError(f"Unable to find pyproject.toml file, got '{project_file}'")
        return project_file

    @computed_field
    @property
    def _project_info_as_dict(self) -> dict:
        """pyproject.toml file as dictionary."""
        return files.TOML(path=self._pyproject_file).data_to_dict()

    @computed_field
    @property
    def _project_info(self) -> Project:
        """Project information."""
        info = Project(name=self._project_info_as_dict["project"]["name"],
                       version=self._project_info_as_dict["project"]["version"],
                       description=self._project_info_as_dict["project"]["description"])
        return info

    @property
    def name(self) -> str:
        """Project name."""
        return self._project_info.name

    @property
    def version(self) -> str:
        """Project name."""
        project_version = self._project_info.version
        return project_version if project_version is not None else "No project version set."

    @property
    def description(self) -> str:
        """Project description."""
        project_description = self._project_info.description
        return project_description if project_description is not None else "No project description given."


@dataclass
class System:
    """System information."""

    os: str

    @field_validator("os")
    @classmethod
    def os_must_be_supported(cls, v: str) -> str:
        """Validate that the operating system is supported."""
        supported_oses = ["Darwin", "Linux", "Windows"]
        if v not in supported_oses:
            raise ValueError(f"Operating system '{v}' is not supported.")
        return v


class SystemSettings(BaseSettings):
    """System settings."""

    @computed_field
    @property
    def _system_info(self) -> System:
        """System information."""
        info = System(os=platform.system())
        return info

    @property
    def os(self) -> str:
        """Operating system."""
        return self._system_info.os


class Token(BaseSettings):
    """Credential details."""

    id: SecretStr



"""Tests for domain.files module."""
import pytest
from bacore.domain import settings

pytestmark = pytest.mark.domain


@pytest.fixture
def fixture_project_root_dir(fixture_pyproject_file):
    return fixture_pyproject_file.parent


class TestCredential:
    """Tests for Credential entity."""

    credential = settings.Credential(username="username", password="passw0rd")

    def test_username(self):
        """Test username."""
        assert self.credential.username == "username"

    def test_username_must_not_contain_spaces(self):
        """Test username must not contain spaces."""
        with pytest.raises(ValueError):
            settings.Credential(username="user name")

    def test_password(self):
        """Test password."""
        assert self.credential.password.get_secret_value() == "passw0rd"

    def test_password_must_be_secret(self):
        """Test password must be of type pydantic.SecretStr."""
        assert isinstance(self.credential.password, settings.SecretStr)


class TestProject:
    """Tests for ProjectInfo entity."""

    def test_name(self):
        """Test name."""
        p = settings.Project(name="bacore")
        assert p.name == "bacore"

    def test_name_must_not_contain_spaces(self):
        """Test name_must_not_contain_spaces."""
        with pytest.raises(ValueError):
            settings.Project(name="ba core")


class TestProjectSettings:
    """Tests for ProjectSettings entity."""

    def test_path_must_be_directory(self, fixture_project_root_dir):
        """Test path_must_be_directory."""
        with pytest.raises(ValueError):
            settings.ProjectSettings(project_root_dir=fixture_project_root_dir / "wrong_path")

    def test_project_settings(self, fixture_project_root_dir):
        """Test ProjectSettings."""
        project_settings = settings.ProjectSettings(project_root_dir=fixture_project_root_dir)
        assert project_settings.name == "bacore"
        assert project_settings.version == "1.0.0"
        assert project_settings.description == "BACore is a framework for business analysis and test automation."


class TestSystem:
    """Test for SystemInfo."""

    def test_os(self):
        """Test os. (Darwin is macOS.)"""
        system_info = settings.System(os="Darwin")
        assert system_info.os in ["Darwin", "Linux", "Windows"]

    def test_os_must_be_supported(self):
        """Test os_must_be_supported."""
        with pytest.raises(ValueError):
            settings.System(os="AS/400")



"""Tests for domain.files module."""
import pytest
from bacore.domain import files
from pathlib import Path

pytestmark = pytest.mark.domain


class TestTOML:
    """Tests for TOML entity."""

    def test_path(self, fixture_pyproject_file):
        """Test path."""
        toml_file = files.TOML(path=fixture_pyproject_file)
        assert isinstance(toml_file.path, Path)

    def test_path_fail_with_string(self):
        """Test path."""
        with pytest.raises(TypeError):
            files.TOML(path="pyproject.toml")

    def test_data_to_dict(self, fixture_pyproject_file):
        """Test toml_file_content."""
        content = files.TOML(path=fixture_pyproject_file)
        assert isinstance(content.data_to_dict(), dict)



"""Module for domain system."""
from shutil import which


class CommandLineProgram:
    """Program."""

    def __init__(self, name: str, verify_exists_func: callable = which):
        """Initialize."""
        self.name = name
        self.verify_exists_func = verify_exists_func

    def verify_exists(self) -> bool:
        """Verify command line program exists on path."""
        return self.verify_exists_func(self.name) is not None



"""Module for handling exceptions."""
from pydantic import ValidationError


class PydValErrInfo:
    """Extract information from pydantic ValidationError."""

    @staticmethod
    def error_msg(e: ValidationError) -> str:
        """Error message from pydantic ValidationError."""
        return e.errors()[0]["ctx"]["error"]

    @staticmethod
    def input(e: ValidationError) -> str:
        """Input value, as it is read by, pydantic ValidationError."""
        return e.errors()[0]["input"]



"""Serve CLI module."""
import subprocess as sup
from bacore.domain import settings
from bacore.domain.errors import PydValErrInfo
from bacore.interfaces import cli_typer
from pydantic import ValidationError
from rich import print
from typer import Argument, Exit, Option, Typer, prompt
from typing import Annotated

app = Typer(rich_markup_mode="rich")


@app.command(rich_help_panel="Serve")
def documentation(project: Annotated[str, Argument(help="Name of project.")] = '',
                  port: Annotated[int, Option(help="Port to serve documentation on.")] = 8000):
    """Serve documentation with MkDocs for a project."""
    cli_typer.verify_programs_installed(['mkdocs'])

    if project == '':
        project = prompt("Enter project name")

    try:
        project = settings.Project(name=project)
    except ValidationError as e:
        print(f'[red]{PydValErrInfo.error_msg(e)}[/red] Input was: "{PydValErrInfo.input(e)}"')
        raise Exit()

    print(f'Serving documentation for project "{project.name}"[white]...[/]')
    sup.run(f"mkdocs serve -a 127.0.0.1:{port} &", shell=True)



"""CLI created with Typer."""
from bacore.domain import files, settings, system
from bacore.interactors import retrieve, verify
from rich import print
from pathlib import Path
from typer import Exit
from typing import Optional


class ProjectInfo:
    """ProjectInfo information."""

    def __init__(self, pyproject_file: Path):
        """Initialize."""
        self._pyproject_file_toml_object = files.TOML(path=pyproject_file)
        self._project_info_dict = retrieve.file_as_dict(file=self._pyproject_file_toml_object)
        self._project_info = settings.Project(name=self._project_info_dict["project"]["name"],
                                          version=self._project_info_dict["project"]["version"],
                                          description=self._project_info_dict["project"]["description"]
                                          )

    @property
    def name(self) -> str:
        """ProjectInfo name."""
        return self._project_info.name

    @property
    def version(self) -> Optional[str]:
        """ProjectInfo version."""
        return self._project_info.version

    @property
    def description(self) -> Optional[str]:
        """ProjectInfo description."""
        return self._project_info.description


def verify_programs_installed(list_of_programs: list[system.CommandLineProgram]):
    """Check if a list of command line programs are installed.

    TODO: This function does not follow the overall design of the project. -> Move to an interactor module.

    """
    programs_not_installed = 0

    for program in list_of_programs:
        if verify.command_on_path(program) is False:
            programs_not_installed += 1
            print(f'{program} is [red]not installed[/]. Install with: [blue]pip install bacore\\[cli\\][/]')

    if programs_not_installed > 0:
        raise Exit(code=1)



"""Module for verification."""
from typing import Protocol


class Exists(Protocol):
    """Protocol for verification of existing."""

    def verify_exists(self) -> bool:
        """Verify if exists."""
        ...


def command_on_path(command: Exists) -> bool:
    """Check if CLI command is on path."""
    return command.verify_exists()



"""Create CLI module."""
import subprocess as sup
from bacore.domain import settings
from bacore.domain.errors import PydValErrInfo
from bacore.interfaces import cli_typer
from pydantic import ValidationError
from rich import print
from typer import Argument, Exit, Typer, prompt
from typing import Annotated

app = Typer(rich_markup_mode="rich")


@app.command(rich_help_panel="Create")
def project(name: Annotated[str, Argument(help="Name of project ([red]no spaces allowed[/]).")] = ''):
    """Create new project ([blue]with hatch[/])."""
    cli_typer.verify_programs_installed(['hatch'])

    if name == '':
        name = prompt("Enter project name")

    try:
        new_project = settings.Project(name=name)
    except ValidationError as e:
        print(f'[red]{PydValErrInfo.error_msg(e)}[/red] Input was: "{PydValErrInfo.input(e)}"')
        raise Exit()

    print(f'Creating new project "{new_project.name}"[white]...[/]')
    sup.run(f'hatch new {new_project.name}', shell=True)



"""BACore CLI entrypoint module."""
from bacore.cli import create, serve
from typer import Typer

app = Typer(rich_markup_mode="rich", add_completion=False)
app.add_typer(create.app, name="create", help="Create a project")
app.add_typer(serve.app, name="serve", help="Serve documentation")
